# coding: utf-8
#

import base64
import io
import json
import os
import traceback

import tornado
from logzero import logger
from PIL import Image
from tornado.escape import json_decode

from ..device import connect_device, get_device
from ..version import __version__


pathjoin = os.path.join


class BaseHandler(tornado.web.RequestHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with")
        self.set_header("Access-Control-Allow-Credentials",
                        "true")  # allow cookie
        self.set_header('Access-Control-Allow-Methods',
                        'POST, GET, PUT, DELETE, OPTIONS')

    def options(self, *args):
        self.set_status(204)  # no body
        self.finish()

    def check_origin(self, origin):
        """ allow cors request """
        return True


class VersionHandler(BaseHandler):
    def get(self):
        ret = {
            'name': "weditor",
            'version': __version__,
        }
        self.write(ret)


class MainHandler(BaseHandler):
    def get(self):
        self.render("index.html")


class DeviceConnectHandler(BaseHandler):
    def post(self):
        platform = self.get_argument("platform").lower()
        device_url = self.get_argument("deviceUrl")

        try:
            id = connect_device(platform, device_url)
        except RuntimeError as e:
            self.set_status(500)
            self.write({
                "success": False,
                "description": str(e),
            })
        except Exception as e:
            logger.warning("device connect error: %s", e)
            self.set_status(500)
            self.write({
                "success": False,
                "description": traceback.format_exc(),
            })
        else:
            ret = {
                "deviceId": id,
                'success': True,
            }
            if platform == "android":
                ws_addr = get_device(id).device.address.replace("http://", "ws://") # yapf: disable
                ret['screenWebSocketUrl'] = ws_addr + "/minicap"
            self.write(ret)


class DeviceHierarchyHandler(BaseHandler):
    def get(self, device_id):
        d = get_device(device_id)
        self.write(d.dump_hierarchy())


class DeviceHierarchyHandlerV2(BaseHandler):
    def get(self, device_id):
        d = get_device(device_id)
        self.write(d.dump_hierarchy2())


class WidgetPreviewHandler(BaseHandler):
    def get(self, id):
        self.render("widget_preview.html", id=id)


class DeviceWidgetListHandler(BaseHandler):
    __store_dir = os.path.expanduser("~/.weditor/widgets")

    def generate_id(self):
        os.makedirs(self.__store_dir, exist_ok=True)
        names = [
            name for name in os.listdir(self.__store_dir)
            if os.path.isdir(os.path.join(self.__store_dir, name))
        ]
        return "%05d" % (len(names) + 1)

    def get(self, widget_id: str):
        data_dir = os.path.join(self.__store_dir, widget_id)
        with open(pathjoin(data_dir, "hierarchy.xml"), "r",
                  encoding="utf-8") as f:
            hierarchy = f.read()

        with open(os.path.join(data_dir, "meta.json"), "rb") as f:
            meta_info = json.load(f)
            meta_info['hierarchy'] = hierarchy
            self.write(meta_info)

    def json_parse(self, source):
        with open(source, "r", encoding="utf-8") as f:
            return json.load(f)

    def put(self, widget_id: str):
        """ update widget data """
        data = json_decode(self.request.body)
        target_dir = os.path.join(self.__store_dir, widget_id)
        with open(pathjoin(target_dir, "hierarchy.xml"), "w",
                  encoding="utf-8") as f:
            f.write(data['hierarchy'])

        # update meta
        meta_path = pathjoin(target_dir, "meta.json")
        meta = self.json_parse(meta_path)
        meta["xpath"] = data['xpath']
        with open(meta_path, "w", encoding="utf-8") as f:
            f.write(json.dumps(meta, indent=4, ensure_ascii=False))

        self.write({
            "success": True,
            "description": f"widget {widget_id} updated",
        })

    def post(self):
        data = json_decode(self.request.body)
        widget_id = self.generate_id()
        target_dir = os.path.join(self.__store_dir, widget_id)
        os.makedirs(target_dir, exist_ok=True)

        image_fd = io.BytesIO(base64.b64decode(data['screenshot']))
        im = Image.open(image_fd)
        im.save(pathjoin(target_dir, "screenshot.jpg"))

        lx, ly, rx, ry = bounds = data['bounds']
        im.crop(bounds).save(pathjoin(target_dir, "template.jpg"))

        cx, cy = (lx + rx) // 2, (ly + ry) // 2
        # TODO(ssx): missing offset
        # pprint(data)
        widget_data = {
            "resource_id": data["resourceId"],
            "text": data['text'],
            "description": data["description"],
            "target_size": [rx - lx, ry - ly],
            "package": data["package"],
            "activity": data["activity"],
            "class_name": data['className'],
            "rect": dict(x=lx, y=ly, width=rx-lx, height=ry-ly),
            "window_size": data['windowSize'],
            "xpath": data['xpath'],
            "target_image": {
                "size": [rx - lx, ry - ly],
                "url": f"http://localhost:17310/widgets/{widget_id}/template.jpg",
            },
            "device_image": {
                "size": im.size,
                "url": f"http://localhost:17310/widgets/{widget_id}/screenshot.jpg",
            },
            # "hierarchy": data['hierarchy'],
        } # yapf: disable

        with open(pathjoin(target_dir, "meta.json"), "w",
                  encoding="utf-8") as f:
            json.dump(widget_data, f, ensure_ascii=False, indent=4)

        with open(pathjoin(target_dir, "hierarchy.xml"), "w",
                  encoding="utf-8") as f:
            f.write(data['hierarchy'])

        self.write({
            "success": True,
            "id": widget_id,
            "note": data['text'] or data['description'],  # 备注
            "data": widget_data,
        })


class DeviceScreenshotHandler(BaseHandler):
    def get(self, serial):
        logger.info("Serial: %s", serial)
        try:
            d = get_device(serial)
            buffer = io.BytesIO()
            d.screenshot().convert("RGB").save(buffer, format='JPEG')
            b64data = base64.b64encode(buffer.getvalue())
            response = {
                "type": "jpeg",
                "encoding": "base64",
                "data": b64data.decode('utf-8'),
            }
            self.write(response)
        except EnvironmentError as e:
            traceback.print_exc()
            self.set_status(500, "Environment Error")
            self.write({"description": str(e)})
        except RuntimeError as e:
            self.set_status(500)  # Gone
            self.write({"description": traceback.format_exc()})




#! /usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import, print_function

import argparse
import base64
import hashlib
import io
import json
import os
import platform
import queue
import signal
import socket
import subprocess
import sys
import time
import traceback
import uuid
import webbrowser
# `pip install futures` for python2
from concurrent.futures import ThreadPoolExecutor
from subprocess import PIPE

import requests
import six
import tornado.escape
import tornado.ioloop
import tornado.web
import tornado.websocket
from logzero import logger
from tornado import gen
from tornado.concurrent import run_on_executor
from tornado.escape import json_encode
from tornado.log import enable_pretty_logging

from .web.handlers.page import (
    BaseHandler, DeviceConnectHandler,
    DeviceHierarchyHandler, DeviceHierarchyHandlerV2, DeviceScreenshotHandler,
    DeviceWidgetListHandler, MainHandler, VersionHandler, WidgetPreviewHandler)
from .web.handlers.proxy import StaticProxyHandler
from .web.handlers.shell import PythonShellHandler
from .web.utils import current_ip, tostr
from .web.version import __version__

enable_pretty_logging()

__dir__ = os.path.dirname(os.path.abspath(__file__))

is_closing = False

if os.name == "nt":
    os.environ["HOME"] = os.path.expanduser("~")
PID_FILEPATH = os.path.expandvars("$HOME/.weditor/weditor.pid")
os.makedirs(os.path.dirname(PID_FILEPATH), exist_ok=True)


def signal_handler(signum, frame):
    global is_closing
    print('exiting...')
    is_closing = True


def stop_server():
    tornado.ioloop.IOLoop.instance().stop()


def try_exit():
    global is_closing
    if is_closing:  # clean up here
        stop_server()
        logger.info('exit success')


class QuitHandler(BaseHandler):
    def get(self):
        stop_server()
        self.write({"success": True, "description": "Successfully quited"})


class CropHandler(BaseHandler):
    def get(self):
        """ used for crop image """
        pass


def make_app(settings={}):
    application = tornado.web.Application(
        [
            (r"/", MainHandler),
            (r"/api/v1/version", VersionHandler),
            (r"/api/v1/connect", DeviceConnectHandler),
            (r"/api/v1/crop", CropHandler),
            (r"/api/v1/devices/([^/]+)/screenshot", DeviceScreenshotHandler),
            (r"/api/v1/devices/([^/]+)/hierarchy", DeviceHierarchyHandler),
            # (r"/api/v1/devices/([^/]+)/exec", DeviceCodeDebugHandler),
            (r"/api/v1/devices/([^/]+)/widget", DeviceWidgetListHandler),
            (r"/api/v1/widgets", DeviceWidgetListHandler),  # add widget
            (r"/api/v1/widgets/([^/]+)", DeviceWidgetListHandler),
            # v2
            (r"/api/v2/devices/([^/]+)/hierarchy", DeviceHierarchyHandlerV2),
            # widgets
            (r"/widgets/([^/]+)", WidgetPreviewHandler),
            (r"/widgets/(.+/.+)", tornado.web.StaticFileHandler, {
                "path": "./widgets"
            }),
            # cache static assets
            (r"/(unpkg.com/.*)", StaticProxyHandler),
            (r"/(cdn.jsdelivr.net/.*)", StaticProxyHandler),
            # (r"/ws/v1/build", BuildWSHandler),
            (r"/ws/v1/python", PythonShellHandler),
            (r"/quit", QuitHandler),
        ],
        **settings)
    return application


def get_running_version(addr: str):
    """
    Returns:
        None if not running
        version string if running
    """
    try:
        r = requests.get(f"{addr}/api/v1/version",
                         timeout=2.0)
        if r.status_code == 200:
            return r.json().get("version", "dev")
    except requests.exceptions.ConnectionError:
        pass
    except Exception as e:
        print("Unknown error: %r" % e)


def cmd_quit(port=17310):
    try:
        requests.get(f"http://127.0.0.1:{port}/quit", timeout=3)
        logger.info("weditor quit successfully")
    except requests.ConnectionError:
        logger.info("weditor already stopped")
    except requests.Timeout:
        logger.info("kill through pid file")
        if not os.path.isfile(PID_FILEPATH):
            logger.warning("Pidfile: %s not exist", PID_FILEPATH)
            return
        
        with open(PIDFILEPATH, "r") as f:
            pid = int(f.read())
            if os.name == "nt": # windows
                subprocess.call(f"taskkill /PID {pid} /T /F") # /F: 强制 /T: 包含子进程
            else:
                os.kill(pid, signal.SIGKILL)
        os.unlink(PID_FILEPATH)
        logger.info("weditor was killed")


def run_web(debug=False, port=17310, open_browser=False, force_quit=False):
    base_url = f"http://localhost:{port}"
    version = get_running_version(base_url)
    if version:
        if force_quit:
            logger.info(f"quit previous weditor server (version: {version})")
            requests.get(base_url + "/quit")
            time.sleep(.5)
        else:
            sys.exit(f"Another weditor({version}) is already running")

    if open_browser:
        webbrowser.open(f'http://localhost:{port}', new=2)

    application = make_app({
        'static_path': os.path.join(__dir__, 'static'),
        'template_path': os.path.join(__dir__, 'templates'),
        'debug': debug,
    })
    print('listening on http://%s:%d' % (current_ip(), port))
    if debug:
        logger.info("enable debug mode")
    signal.signal(signal.SIGINT, signal_handler)
    application.listen(port)

    with open(PID_FILEPATH, "w") as f:
        f.write(str(os.getpid()))

    tornado.ioloop.PeriodicCallback(try_exit, 100).start()
    tornado.ioloop.IOLoop.instance().start()
    # tornado.ioloop.IOLoop.instance().add_callback(consume_queue)

    if os.path.exists(PID_FILEPATH):
        os.unlink(PID_FILEPATH)


def create_shortcut():
    if os.name != 'nt':
        sys.exit("Shortcut only available in Windows")

    import pythoncom  # pyint: disable=import-error
    from win32com.shell import shell
    from win32com.shell import shellcon
    # Refs
    # - https://github.com/pearu/iocbio/blob/master/installer/utils.py
    # - https://blog.csdn.net/thundor/article/details/5968581
    ilist = shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_DESKTOP)
    dtpath = shell.SHGetPathFromIDList(ilist).decode('utf-8')

    shortcut = pythoncom.CoCreateInstance(shell.CLSID_ShellLink, None,
                                          pythoncom.CLSCTX_INPROC_SERVER,
                                          shell.IID_IShellLink)
    launch_path = sys.executable
    shortcut.SetPath(launch_path)
    shortcut.SetArguments("-m weditor")
    shortcut.SetDescription(launch_path)
    shortcut.SetIconLocation(sys.executable, 0)
    shortcut.QueryInterface(pythoncom.IID_IPersistFile).Save(
        dtpath + "\\WEditor.lnk", 0)
    print("Shortcut created. " + dtpath + "\\WEditor.lnk")


def main():
    # yapf: disable
    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument("-v", "--version", action="store_true", help="show version")
    ap.add_argument('-q', '--quiet', action='store_true', help='quite mode, no open new browser')
    ap.add_argument('-p', '--port', type=int, default=17310, help='local listen port for weditor')
    ap.add_argument("-f", "--force-quit", action='store_true', help="force quit before start")
    ap.add_argument('--debug', action='store_true', help='open debug mode')
    ap.add_argument('--shortcut', action='store_true', help='create shortcut in desktop')
    ap.add_argument("--quit", action="store_true", help="stop weditor")
    args = ap.parse_args()
    # yapf: enable

    if args.version:
        print(__version__)
        return

    if args.shortcut:
        create_shortcut()
        return
    
    if args.quit:
        cmd_quit(args.port)
        return

    if sys.platform == 'win32' and sys.version_info[:2] >= (3, 8):
        import asyncio
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    open_browser = not args.quiet and not args.debug
    run_web(args.debug, args.port, open_browser, args.force_quit)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Thu Mar 30 2023 17:03:48 by codeskyblue
"""

import pathlib
import re
import pytest
from weditor.web import utils

@pytest.fixture
def hello_path(tmp_path: pathlib.Path) -> pathlib.Path:
    test_path = tmp_path / 'test.txt'
    test_path.write_text('hello')
    return test_path


def test_tostr():
    assert utils.tostr('hello') == "hello"


def test_read_file_content(hello_path: pathlib.Path):
    assert utils.read_file_content(hello_path) == b"hello"


def test_sha_file(hello_path: pathlib.Path):
    assert utils.sha_file(hello_path) == "aaf4c61ddcc5e8a2dabede0f3b482cd9aea9434d"


def test_write_file_content(hello_path: pathlib.Path):
    utils.write_file_content(hello_path, 'world')
    assert hello_path.read_text() == "world"


def test_current_ip():
    ip = utils.current_ip()
    assert re.match(r"\d+\.\d+\.\d+\.\d+", ip)


# coding: utf-8
#
import asyncio
import atexit
import json
import logging
import os
import re
import signal
import subprocess
import sys
import tempfile
import threading
from typing import Any

import tornado.iostream
import tornado.queues
import tornado.websocket
from tornado import gen
from tornado.ioloop import IOLoop
from tornado.process import Subprocess

logger = logging.getLogger("weditor")
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
IS_WINDOWS = os.name == "nt"


class WinAsyncSubprocess(object):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.io_loop = IOLoop.current()
        kwargs['bufsize'] = 0  # unbuffed
        kwargs['stdin'] = subprocess.PIPE
        kwargs['stdout'] = subprocess.PIPE
        kwargs['stderr'] = subprocess.STDOUT

        self.proc = subprocess.Popen(*args, **kwargs)
        self.pid = self.proc.pid

        # https://www.tornadoweb.org/en/stable/queues.html
        self._qout = tornado.queues.Queue()
        self._qexit = tornado.queues.Queue()
        threading.Thread(name="async-subprocess",
                         target=self._drain,
                         daemon=True).start()

    def _drain(self):
        logger.info("Started drain subprocess stdout in thread")
        for line in iter(self.proc.stdout.readline, b''):
            self.io_loop.add_callback(self._qout.put, line)
        self.io_loop.add_callback(self._qout.put, None)
        logger.info("windows process stdout closed")
        self.io_loop.add_callback(self._qexit.put, self.proc.wait())

    async def wait_for_exit(self, raise_error=True):
        """ Ignore raise_error """
        exit_code = await self._qexit.get()
        return exit_code

    async def readline(self) -> bytes:
        ret = await self._qout.get()
        if ret is None:
            raise IOError("subprocess stdout closed")
        return ret

    async def stdin_write(self, data: bytes):
        return self.proc.stdin.write(data)


class PosixAsyncSubprocess(Subprocess):
    async def readline(self) -> bytes:
        return await self.stdout.read_until(b"\n")

    async def stdin_write(self, data: bytes):
        return await self.stdin.write(data)


class PythonShellHandler(tornado.websocket.WebSocketHandler):
    def initialize(self):
        self._tmpd = tempfile.TemporaryDirectory(suffix='-weditor')
        atexit.register(self._tmpd.cleanup)

    def on_close(self):
        logger.warning("websocket closed, cleanup")
        self._tmpd.cleanup()
        atexit.unregister(self._tmpd.cleanup)
        IOLoop.current().add_callback(self.kill_process)

    @property
    def _tmpdir(self) -> str:
        return self._tmpd.name

    async def prepare(self):
        """
        Refs:
            https://www.tornadoweb.org/en/stable/process.html#tornado.process.Subprocess
            https://www.tornadoweb.org/en/stable/iostream.html#tornado.iostream.IOStream
        """
        AsyncSubprocess = WinAsyncSubprocess if IS_WINDOWS else PosixAsyncSubprocess
        env = os.environ.copy()
        env['PYTHONIOENCODING'] = "utf-8"
        self.__process = AsyncSubprocess(
            [sys.executable, "-u",
             os.path.join(ROOT_DIR, "../ipyshell-console.py")],
            env=env,
            cwd=self._tmpdir,
            stdin=Subprocess.STREAM,
            stdout=Subprocess.STREAM,
            stderr=subprocess.STDOUT) # yapf: disable
        # self.__process = Subprocess([sys.executable, "-u", os.path.join(ROOT_DIR, "ipyshell.py")],
        #     # bufsize=1, #universal_newlines=True,
        #     stdin=Subprocess.STREAM,
        #     stdout=Subprocess.STREAM)
        IOLoop.current().add_callback(self.sync_process_output)

    async def kill_process(self):
        self.__process.proc.kill()
        ret = await self.__process.wait_for_exit(raise_error=False)
        logger.info("process quited with code %d", ret)

    async def _readline_decoded(self) -> str:
        line = await self.__process.readline()
        return line.decode("utf-8").rstrip()

    async def sync_process_output(self):
        try:
            while True:
                line = await self._readline_decoded()
                if not line:
                    logger.warning("proc-stdout read empty line")
                    break
                fields = line.split(":", 1)
                if len(fields) != 2:
                    continue
                cmdx, value = fields
                if cmdx == "LNO":
                    self.write2({"method": "gotoLine", "value": int(value)})
                elif cmdx == "DBG":
                    logger.debug("DBG: %s", value)
                    # self.write2({"method": "output", "value": "- "+value})
                elif cmdx == "WRT":
                    # here value is json_encoded string
                    self.write2({
                        "method": "output",
                        "value": json.loads(value)
                    })
                elif cmdx == "EOF":
                    logger.debug(
                        "finished running code block, time used %.1fs",
                        int(value) / 1000)
                    self.write2({"method": "finish", "value": int(value)})
                else:
                    # self.write2({"method": "output", "value": line +"\n"})
                    logger.warning("Unsupported output line: %s", line)
        except (tornado.iostream.StreamClosedError, IOError):
            pass
        finally:
            logger.debug("sync process output stopped")
            # code may never goes here
            #ret = await self.__process.wait_for_exit(raise_error=False)
            #logger.info("process exit with code %d", ret)
            #self.__process = None

    def send_keyboard_interrupt(self):
        if IS_WINDOWS:  # Windows
            # On windows, it's not working with the following code
            # - p.send_signal(signal.SIGINT)
            # - os.kill(p.pid, signal.CTRL_C_EVENT)
            # - subprocess.call(["taskkill", "/PID", str(p.pid)])
            # But the following command works find
            pid = self.__process.pid
            import ctypes
            k = ctypes.windll.kernel32

            k.FreeConsole()  # Don't understand
            k.AttachConsole(pid)
            k.SetConsoleCtrlHandler(
                None, True)  # Disable Ctrl-C handling for our program
            k.GenerateConsoleCtrlEvent(signal.CTRL_C_EVENT, 0)  # SIGINT

            # Re-enable Ctrl-C handling or any subsequently started
            # programs will inherit the disabled state.
            k.SetConsoleCtrlHandler(None, False)
        else:
            self.__process.proc.send_signal(
                signal.SIGINT)  # Linux is so simple

    async def open(self):
        logger.debug("websocket opened")
        logger.info("create process pid: %d", self.__process.pid)
        # self.write2({"method": "resetContent", "value": INIT_CODE})
        # self.write2({"method": "gotoLine", "value": 1})
        # await gen.sleep(.1)

    def write2(self, data):
        self.write_message(json.dumps(data))

    def _adjust_code(self, code: str):
        """ fix indent error, remove all line spaces """
        prefixs = re.findall(r"^\s*", code, re.M)
        space_len = min([len(pre) for pre in prefixs])
        lines = code.splitlines(keepends=True)
        return ''.join([line[space_len:] for line in lines])

    async def on_message(self, message):
        # print("Receive:", message)
        data = json.loads(message)
        method, value = data['method'], data.get('value')
        if method == 'input':
            code = self._adjust_code(value)
            code = json.dumps(code) + "\n"
            logger.debug("send to proc: %s", code.rstrip())
            await self.__process.stdin_write(code.encode('utf-8'))
        elif method == "keyboardInterrupt":
            self.send_keyboard_interrupt()
        elif method == "restartKernel":
            await self.kill_process()
            await self.prepare()
            self.write2({"method": "restarted"})
        else:
            logger.warning("Unknown received message: %s", data)
